/*
 * Copyright (c) 2005 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ident "$Id$"

#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/module.h>
#include <linux/pci.h>

#include <asm/uaccess.h>

MODULE_AUTHOR("Matthieu Herrb");
MODULE_DESCRIPTION("Contec PIO32/32L PCI DIO Board driver");
MODULE_LICENSE("BSD");

#include "contec_dio.h"

#define CONTEC_DIO_NAME "contec_dio"

#define PCI_VENDOR_ID_CONTEC 0x1221
#define PCI_DEVICE_ID_PIO3232L 0x8122

struct dio_dev {
	int usecount;
	unsigned long baseaddr;
	unsigned long size;
	struct cdev cdev;
};

/**
 ** Device numbers - default do dynamic assignement 
 **/
#ifndef CONTEC_DIO_MAJOR
# define CONTEC_DIO_MAJOR 0
#endif

#ifndef CONTEC_DIO_MINOR
# define CONTEC_DIO_MINOR 0
#endif

#ifdef DEBUG_CONTEC_DIO
#define debug_print(x) printk x
#else
#define debug_print(x) 
#endif

int devmajor = CONTEC_DIO_MAJOR;
int devminor = CONTEC_DIO_MINOR;

module_param(devmajor, int, S_IRUGO);
module_param(devminor, int, S_IRUGO);

struct dio_dev dio;

static int
contec_dio_probe(struct pci_dev *dev, const struct pci_device_id *id)
{
	unsigned long baseaddr, endaddr;
	struct resource *res;
	
	printk(KERN_INFO "%s: found device (%04X,%04X) on %s\n",
	    CONTEC_DIO_NAME, PCI_VENDOR_ID_CONTEC, PCI_DEVICE_ID_PIO3232L,
	    pci_name(dev));
	if (pci_enable_device(dev) < 0) {
		printk(KERN_WARNING "%s: error enabling device\n",
		    CONTEC_DIO_NAME);
		return -1;
	}
	baseaddr = pci_resource_start(dev, 0);
	endaddr = pci_resource_end(dev, 0);
	res = request_region(baseaddr, endaddr - baseaddr + 1,
	    CONTEC_DIO_NAME);
	printk(KERN_INFO "\tio addr=0x%lx len %lu\n", baseaddr, endaddr - baseaddr + 1);
	dio.baseaddr = baseaddr;
	dio.size = endaddr - baseaddr + 1;
	return 0;
}

static void
contec_dio_remove(struct pci_dev *dev)
{
	release_region(dio.baseaddr, dio.size);
	dio.baseaddr = 0;
	return;
}

/*----------------------------------------------------------------------*/

/*
 * file operations
 */
static int
contec_dio_open(struct inode *inode, struct file *file)
{
	if (dio.usecount != 0)
		return -EBUSY;
	dio.usecount++;
	return 0;
}

static int
contec_dio_release(struct inode *inode, struct file *file)
{
	dio.usecount--;
	return 0;
}

static int
contec_dio_ioctl(struct inode *inode, struct file *file,
    unsigned int function, unsigned long arg)
{
	DIO_ARG parg;

	if (!access_ok(VERIFY_WRITE, (void *)arg, sizeof(DIO_ARG))) {
		printk(KERN_NOTICE "%s verify error\n", __func__);
		return -EFAULT;
	}
	if (copy_from_user(&parg, (void *)arg, sizeof(DIO_ARG))) {
		printk(KERN_NOTICE "%s copy_from_user error\n", __func__);
		return -EFAULT;
	}
	switch (function) {
	case CONTEC_DIO_READ:
		rmb();
		parg.value = inb(dio.baseaddr + parg.addr);
		if (copy_to_user((void *)arg, (void *)&parg, sizeof(DIO_ARG)))
			return -EFAULT;
		debug_print((KERN_INFO "%s read addr %x value %x\n",
			     __func__, parg.addr, parg.value));
		break;

	case CONTEC_DIO_WRITE:
		outb(parg.value, dio.baseaddr + parg.addr);
		wmb();
		debug_print((KERN_INFO "%s write addr %x value %x\n",
			     __func__, parg.addr, parg.value));
		break;

	default:
		printk(KERN_NOTICE "%s unknown function %d\n", 
		       __func__, function);
		return -EINVAL;

	}
	return 0;
}

/*----------------------------------------------------------------------*/

/**
 ** module glue
 **/

static const struct pci_device_id contec_dio_ids[] = {
	{ PCI_DEVICE(PCI_VENDOR_ID_CONTEC, PCI_DEVICE_ID_PIO3232L), },
	{ 0, },
};

MODULE_DEVICE_TABLE(pci, contec_dio_ids);

static struct pci_driver pci_driver = {
	.name = CONTEC_DIO_NAME,
	.id_table = contec_dio_ids,
	.probe = contec_dio_probe,
	.remove = contec_dio_remove,
};

struct file_operations contec_dio_fops = {
	.open = contec_dio_open,
	.release = contec_dio_release,
	.ioctl = contec_dio_ioctl,
};

static void __exit
contec_dio_cleanup(void)
{
	dev_t dev = MKDEV(devmajor, devminor);

	pci_unregister_driver(&pci_driver);
	cdev_del(&dio.cdev);
	unregister_chrdev_region(dev, 1); /* XXX */
	printk(KERN_INFO "%s: bye\n", CONTEC_DIO_NAME);
}

static int __init
contec_dio_init(void)
{
	int err;
	dev_t dev;

	printk(KERN_INFO "%s: initializing\n", CONTEC_DIO_NAME);
	if ((err = pci_register_driver(&pci_driver)) < 0) 
		return err;
	if (devmajor != 0) {
		dev = MKDEV(devmajor, devminor);
		err = register_chrdev_region(dev, 1, CONTEC_DIO_NAME);
	} else {
		err = alloc_chrdev_region(&dev, 0, 1, CONTEC_DIO_NAME);
		devmajor = MAJOR(dev);
	}
	if (err < 0) {
		printk(KERN_WARNING "%s: can't get major %d\n",
		    CONTEC_DIO_NAME, devmajor);
		return err;
	}
	printk(KERN_INFO "%s: using major number %d\n",
	    CONTEC_DIO_NAME, devmajor);
	cdev_init(&dio.cdev, &contec_dio_fops);
	dio.cdev.owner = THIS_MODULE;
	dio.cdev.ops = &contec_dio_fops;
	if ((err = cdev_add(&dio.cdev, dev, 1)) < 0) {
		printk(KERN_NOTICE "Error %d adding contec dio\n", err);
		return err;
	}
	return 0;
}

module_init(contec_dio_init);
module_exit(contec_dio_cleanup);

/* That's all folks */

